%Bayes, Markov, and Ito:  Linear operators for control of stochastic
%systems
%
%  Terence Sanger
%
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Damped pendulum model shown in figures 1 and 2.
%
% model is \dot(p)(v) = Lp(v)
%          \dot(p)(x) = Lp(x)
%
%  dv = - (g/l) sin(x) dt - (b/m) v dt + dB1
%  dx = v dt + dB2
%       dB1, dB2 are independent white noise with variance sigma_x and
%       sigma_v 
% this model calculates the joint density p(v,x) assuming independence
%    Therefore 2-step calculation: first the conditional densities  p(x|v), then p(v|x)
%

npoints = 100;  %probability quantization
ntimes = 1000;  %time points.  Total time = ntimes * dt sec.

%values for x and v
min_x = -pi/2;
max_x = pi/2;
min_v = -1;  %rad/sec
max_v = 1;
x = linspace(min_x, max_x, npoints)';
v = linspace(min_v, max_v, npoints)';
dx = (max_x - min_x) / npoints;     %spatial constants for integration
dv = (max_v - min_v) / npoints;

%initial conditions
pvx = zeros(npoints,npoints);           %p(v,x);
pvx(npoints/2,npoints/4:3*npoints/4) = 2/npoints; %zero velocity, uniform distribution of position
pvx = pvx / sum(sum(pvx));    %normalize

%simulation parameters
stiffness = 0;     
damping = 0.4;     
sigma_v = .1;                %noise variance
sigma_x = .1;
dt = 0.01;                   %precision for integration
mass = 1;
length = 5;                  %length of pendulum
damping_ratio = damping / (2*sqrt(stiffness * mass));
gravity = 10;

%pendulum equation  f = -g sin(x)
k = - (gravity / length) * sin(x);   %stiffness term converted from force to acceleration
b = damping * v / mass;  %damping 

%create storage for density at each point in time
Vdensity = zeros(npoints, ntimes);
Xdensity = zeros(size(Vdensity));
clear F;            %movie 


figure
for t=1:ntimes,
    
    %first calculate p(v|x=xi) for all xi
    gv = sigma_v * ones(size(v));            %diffusion term; variance is same for all xi 
    for xi = 1:npoints,
        fv = k(xi) * ones(size(v)) - b;      %p(vdot | x=x_i);  damping is same for all xi
        dpv = Lp(pvx(:,xi), fv/dv, gv/dv);   %calculate and apply the linear Fokker-Planck operator
        pvx(:,xi) = pvx(:,xi) + dpv * dt;    %integrate the conditional probability density
    end
    
    %now, calculate p(x|v=vi) for all vi
    gx = sigma_x * ones(size(x));            %diffusion term; variance is same for all vi
    for vi = 1:npoints,
        fx = v(vi) * ones(size(x));          %velocity update
        dpx = Lp(pvx(vi,:)', fx/dx, gx/dx);  %calculate and apply the linear Fokker-Planck operator
        pvx(vi,:) = pvx(vi,:) + dpx' * dt;   %integrate the conditional probability density
    end

    %maintain normalization and fix small round-off errors
    pvx(pvx<0) = 0;
    pvx = pvx / sum(sum(pvx));
    
    %save marginals for display
    Xdensity(:,t) = sum(pvx);
    Vdensity(:,t) = sum(pvx');
    
    %create movie of the joint density p(v,x)
    if ~mod(t,10)
        imagesc(pvx*512);
        colormap('gray');
        F(t/10) = getframe;
    end
end

%
%  figure 1
%
scrsz = get(0,'ScreenSize');
figure('Position',[scrsz(3)/2 scrsz(4)/2 scrsz(3)/2 2*scrsz(3)/10], 'PaperPositionMode', 'auto')
for i=1:10,
    r = ceil(i/5);
    c = i - 5 * (r-1);
    subplot('position', [0.2*(c-1)+.01 0.5*(2-r)+0.03 0.18 0.4]);
    imagesc(frame2im(F(5+10*(i-1))));
    axis square
    axis tight
    axis off
    title([num2str(i) ' sec'], 'fontsize', 12, 'fontweight', 'bold');
end

%
%  figure 2
%
figure('Position',[scrsz(3)/2 1 scrsz(3)/2 scrsz(3)/4], 'PaperPositionMode', 'auto')
subplot(1,2,1)
image(Xdensity*512);
colormap('gray'); 
xlabel('time (sec)', 'fontsize', 12, 'fontweight', 'bold');
ylabel('position (rad)', 'fontsize', 12, 'fontweight', 'bold');
set(gca, 'XTick', 0:ntimes/2:ntimes);
set(gca, 'XTickLabel', {0, ntimes/200, ntimes/100});
set(gca, 'YTick', [1 npoints/2 npoints]);
set(gca, 'YTickLabel', {'-pi/2', 0, 'pi/2'});


subplot(1,2,2)
image(Vdensity*1024);
colormap('gray'); 
xlabel('time (sec)', 'fontsize', 12, 'fontweight', 'bold');
ylabel('velocity (rad/sec)', 'fontsize', 12, 'fontweight', 'bold');
set(gca, 'XTick', 0:ntimes/2:ntimes);
set(gca, 'XTickLabel', {0, ntimes/200, ntimes/100});
set(gca, 'YTick', [1 npoints/2 npoints]);
set(gca, 'YTickLabel', {'-1', 0, '1'});




